#include <cmath>
#include <cstdio>
#include <vector>
#include <iostream>
#include <algorithm>
#include <memory>
#include <complex>
#include "main_helper.hpp"
#include "scoped_set_adder.hpp"
#include "ui/widget.hpp"
#include "../include/my_lib.hpp"
#include "thread_with_lambda.hpp"
#include "thread_manager.hpp"
#include "deadlock/deadlock.hpp"
 
using namespace std;
using namespace MY_LIB;
 
int length_of_array;

ostream & operator << (ostream & ost, const int a[])
{
    // bad form: using global length_of_array
    for(int i = 0; i < length_of_array; ++i) {
        ost << a[i] << " ";
    }
    return ost;
}
 
// remove all instances of element n, return new length
int remove_element(int a[], int n, int elem)
{
    int new_length = 0;
    for(int i = 0; i < n; ++i) {
        if(a[i] != elem) {
            if(i != new_length) {
                a[new_length] = a[i];
            }
            ++new_length;
        }
    }
    return new_length;
}
 
// remove all duplicates of an element, return new length
int remove_duplicates(int a[], int n)
{
    int last = 0;
    for(int i = 1; i < n; ++i) {
        if(a[last] != a[i]) {
            ++last;
            a[last] = a[i];
        }
    }
    return last + 1;
}

void test_cpp14() {
    // Store a generalized lambda, that squares a number, in a variable
    auto func = [](auto input) { return input * input; };

    // square of an int
    std::cout << func(10) << std::endl;

    // square of a double
    std::cout << func(2.345) << std::endl;

    // square of a complex number
    std::cout << func(std::complex<double>(3, -2)) << std::endl;
}
 
int main() {
    length_of_array = 6;
    int a[6] = { 10, 20, 20, 30, 40, 40 };
    int item_to_remove = 20;

    // thread_with_lambda();

    cout << "Removing " << item_to_remove << endl;
    length_of_array = remove_element(a, length_of_array, item_to_remove);
    cout << "a became " << a << endl;
 
    length_of_array = 6;
    int b[6] = { 10, 20, 20, 30, 40, 40 };
    length_of_array = remove_duplicates(b, length_of_array);
    cout << "Removing duplicates yielded " << length_of_array << " items " << b << endl;

	main_helper_fn();
	UI::widget_cb();
	cout << "sizeof(int) is " << sizeof(int) << "." << endl;
	my_lib_fn();
	my_lib_fn(5);
    my_lib_rect r(16, 9);
	cout << "area of rect r is " << r.area() << "." << endl;
    test_cpp14();

	{
		std::unique_ptr<int> up( new int() );
		scoped_set_adder set_adder("a set_adder object");
		scoped_set_adder another_set_adder("and another set_adder");
		if (up) {
		    cout << "unique_ptr passes if." << endl;
		} else {
		    cout << "unique_ptr fails if." << endl;
		}
		up.release();
		if (up) {
		    cout << "unique_ptr after release passes if." << endl;
		} else {
		    cout << "unique_ptr after release fails if." << endl;
		}
		if (up.get()) {
		    cout << "unique_ptr.get() after release passes if." << endl;
		} else {
		    cout << "unique_ptr.get() after release fails if." << endl;
		}
        scoped_set_adder::dump();
	}

    {   // Testing Herb Sutter's pImpl template
        std::unique_ptr<MyClass> my_class(std::make_unique<MyClass>());
        my_class->f();
    }

    // cout << "deadlock() returned " << deadlock() << endl;

    // This_function_is_not_defined();
 
    cout << "Done." << endl;
    return 0;
}
